from aiohttp import web
import json
from logging import getLogger
_log = getLogger(__name__)


_instance = None


def get_websockets():
    global _instance
    return _instance


def setup_websockets(app=None, options=None):
    global _instance
    if _instance is not None:
        del _instance
    _instance = Websockets(app=app, options=options)
    return _instance


class Websockets():
    def __init__(self, app=None, options=None):
        global _instance
        if _instance is None:
            _instance = self

        self._options = options

        self._ws_list = []
        self._subscriptions = {}

        app.router.add_get(self._options.websocket_path, self._ws_handler)
        app.on_shutdown.append(self.shutdown)

    async def send_to_all(self, msg: dict) -> None:
        for ws in self._ws_list:
            await ws.send_str(msg)
        return None

    def subscribe(self, model: str, callback: callable) -> None:
        """Install a callback function for given model.

        :param model: name of data model that asks for websocket
        :type model: str
        :param callback: callback function that is called when data arrives
        :type callback: function with parameters msg
        :return: None
        :rtype: None
        """
        self._subscriptions[model] = callback
        return None

    async def _ws_handler(self, request) -> web.WebSocketResponse:
        ws = web.WebSocketResponse()
        await ws.prepare(request)

        self._ws_list.append(ws)

    # Here possiblity to send init-message to Web-Client

        try:
            async for msg in ws:
                _log.debug(msg)
                if msg.type == web.WSMsgType.TEXT:
                    msg = json.loads(msg.data)
                    await self._dispatch_message(msg)
                else:
                    break
        except Exception as e:
            _log.error(f'websocket.py, exeption {e}')
        finally:
            _log.debug(f'Finally ws remove: {ws}')
            self._ws_list.remove(ws)
        return ws

    async def _dispatch_message(self, msg: dict) -> bool:
        model = msg.get('scope', None)
        if model is None:
            _log.error(f'Message not for a scope: {msg}')
            return False
        callback_func = self._subscriptions.get(model, None)
        if callback_func is None:
            _log.error(f'Message for an unknown scope: {msg}')
            return False
        try:
            await callback_func(msg)
        except Exception as e:
            _log.error(
                f'websocket.py, exeption {e} in callable {callback_func}')
            return False
        return True

    async def shutdown(self, app) -> None:
        for ws in self._ws_list:
            _log.debug(f'calling ws.close for: {ws}')
            await ws.close()
        return None



import asyncio
import configargparse
import pkg_resources
import os
from pathlib import Path
import socket
from typing import Dict, Any

import logging
from logging.handlers import RotatingFileHandler

from riego.db import setup_db
from riego.mqtt import setup_mqtt
from riego.boxes import setup_boxes
from riego.valves import setup_valves
from riego.timer import setup_timer
from riego.model.parameters import setup_parameters


from riego.web.websockets import setup_websockets
from riego.web.routes import setup_routes
from riego.web.error_pages import setup_error_pages


from aiohttp import web
import jinja2
import aiohttp_jinja2
import aiohttp_debugtoolbar

import base64
from cryptography import fernet
from aiohttp_session import setup as setup_session, get_session
from aiohttp_session.cookie_storage import EncryptedCookieStorage


from riego import __version__


async def alert_ctx_processor(request: web.Request) -> Dict[str, Any]:
    # Jinja2 context processor
    session = await get_session(request)
    alert = session.get("alert")
    session['alert'] = None
    session.changed()
    return {"alert": alert}


def main():
    options = _get_options()

    _setup_logging(options=options)

    try:
        with open('riego.conf', 'xt') as f:
            for item in vars(options):
                f.write(f'# {item}={getattr(options, item)}\n')
    except IOError:
        pass

    if options.defaults:
        for item in vars(options):
            print(f'# {item}={getattr(options, item)}')
        exit(0)

    if options.version:
        print('Version: ', __version__)
        exit(0)

#    if sys.version_info >= (3, 8):
#        asyncio.DefaultEventLoopPolicy = asyncio.WindowsSelectorEventLoopPolicy  # noqa: E501

    if os.name == "posix":
        import uvloop  # pylint: disable=import-error
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    app = web.Application()

    app.on_startup.append(on_startup)
    app.on_shutdown.append(on_shutdown)
    app.on_cleanup.append(on_cleanup)

    app['version'] = __version__
    app['options'] = options

    websockets = setup_websockets(app=app, options=options)
    db = setup_db(options=options)
    parameters = setup_parameters(app=app, options=options, db=db)
    mqtt = setup_mqtt(app=app, options=options)
    setup_boxes(options=options, db=db, mqtt=mqtt)
    valves = setup_valves(options=options, db=db,
                          mqtt=mqtt, websockets=websockets)
    setup_timer(app=app, options=options, db=db,
                mqtt=mqtt, valves=valves, parameters=parameters)

    fernet_key = fernet.Fernet.generate_key()
    secret_key = base64.urlsafe_b64decode(fernet_key)
    setup_session(app, EncryptedCookieStorage(secret_key))

    loader = jinja2.FileSystemLoader(options.http_server_template_dir)
    aiohttp_jinja2.setup(app,
                         loader=loader,
                         # enable_async=True,
                         # context_processors=[alert_ctx_processor],
                         )

    setup_routes(app)
    setup_error_pages(app)

    if options.enable_aiohttp_debug_toolbar:
        aiohttp_debugtoolbar.setup(
            app, check_host=False, intercept_redirects=False)

    logging.getLogger(__name__).info("Start")

    web.run_app(app,
                host=options.http_server_bind_address,
                port=options.http_server_bind_port)


def _setup_logging(options=None):
    formatter = logging.Formatter(
        "%(asctime)s;%(levelname)s;%(name)s;%(message)s ")
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    if options.verbose:
        level = logging.DEBUG
    else:
        level = logging.INFO
    file_handler = RotatingFileHandler(options.log_file, mode='a',
                                       maxBytes=options.log_max_bytes,
                                       backupCount=options.log_backup_count,
                                       encoding=None, delay=0)
    file_handler.setFormatter(formatter)
    Path(options.log_file).parent.mkdir(parents=True, exist_ok=True)
    logging.basicConfig(level=level, handlers=[stream_handler, file_handler])

    logging.getLogger("gmqtt").setLevel(logging.ERROR)
    logging.getLogger("aiohttp.access").setLevel(logging.ERROR)


def _get_options():
    p = configargparse.ArgParser(
        default_config_files=['/etc/riego/conf.d/*.conf', '~/.riego.conf',
                              'riego.conf'])
    p.add('-c', '--config', is_config_file=True, env_var='RIEGO_CONF',
          required=False, help='config file path')
    p.add('-d', '--db_filename', help='Path and name for DB file',
          default='db/riego.db')
    p.add('--db_migrations_dir',
          help='path to database migrations directory',
          default=pkg_resources.resource_filename('riego', 'migrations'))
    p.add('-l', '--log_file', help='Full path to logfile',
          default='log/riego.log')
    p.add('--log_max_bytes', help='Maximum Evet Log Size in bytes',
          default=1024*300, type=int)
    p.add('--log_backup_count', help='How many files to rotate',
          default=3, type=int)
    p.add('-m', '--mqtt_host', help='IP adress of mqtt host',
          default='127.0.0.1')
    p.add('-p', '--mqtt_port', help='Port of mqtt service',
          default=1883, type=int)
    p.add('--mqtt_client_id', help='Client ID for MQTT-Connection',
          default=f'riego_ctrl_{socket.gethostname()}')
    p.add('--mqtt_subscription_topic', help='MQTT Topic that we are listening',
          default='riego/#')
    p.add('--base_dir', help='Change only if you know what you are doing',
          default=Path(__file__).parent)
    p.add('--http_server_bind_address',
          help='http-server bind address', default='0.0.0.0')
    p.add('--http_server_bind_port', help='http-server bind port',
          default=8080, type=int)
    p.add('--http_server_static_dir',
          help='Serve static html files from this directory',
          default=pkg_resources.resource_filename('riego.web', 'static'))
    p.add('--http_server_template_dir',
          help='Serve template files from this directory',
          default=pkg_resources.resource_filename('riego.web', 'templates'))
    p.add('--websocket_path', help='url path for websocket',
          default="/ws")
    p.add('--time_format', help='Store and display time',
          default="%Y-%m-%d %H:%M:%S")
    p.add('--mqtt_cmnd_prefix', help='',
          default="cmnd")
    p.add('--mqtt_result_subscription', help='used by class valves',
          default="stat/+/RESULT")

    p.add('--mqtt_lwt_subscription', help='used by class boxes',
          default="tele/+/LWT")
    p.add('--mqtt_state_subscription', help='used by class boxes',
          default="tele/+/STATE")
    p.add('--mqtt_info1_subscription', help='used by class boxes',
          default="tele/+/INFO1")
    p.add('--mqtt_info2_subscription', help='used by class boxes',
          default="tele/+/INFO2")

    p.add('--mqtt_sensor_subscription', help='yet not used',
          default="tele/+/SENSOR")

    p.add('--mqtt_keyword_ON', help='',
          default="ON")
    p.add('--mqtt_keyword_OFF', help='',
          default="OFF")

    p.add('--parameters_start_time_1', default="19:00")
    p.add('--parameters_max_duartion', default="240")


    p.add('--enable_aiohttp_debug_toolbar', action='store_true')
    p.add('--enable_asyncio_debug', action='store_true')
    p.add('--enable_timer_dev_mode', action='store_true')

    p.add('-v', '--verbose', help='verbose', action='store_true')
    p.add('--version', help='Print version and exit', action='store_true')
    p.add('--defaults', help='Print options with default values and exit',
          action='store_true')

    options = p.parse_args()
    if options.verbose:
        print(p.format_values())

    return options


async def on_startup(app):
    logging.getLogger(__name__).debug("on_startup")
    if app['options'].enable_asyncio_debug:
        asyncio.get_event_loop().set_debug(True)


async def on_shutdown(app):
    logging.getLogger(__name__).debug("on_shutdown")


async def on_cleanup(app):
    logging.getLogger(__name__).debug("on_cleanup")



import aiohttp_jinja2
from aiohttp import web

from riego.db import get_db


router = web.RouteTableDef()


def setup_routes_events(app):
    app.add_routes(router)


@router.get("/events")
@aiohttp_jinja2.template('events/index.html')
async def system_event_log(request):
    c = get_db().conn.cursor()
    c.execute('''SELECT events.*, valves.name
                FROM events, valves
                WHERE events.valve_id = valves.id
                ORDER BY events.created_at DESC''')
    items = c.fetchall()
    get_db().conn.commit()
    return {'items': items}



"""

"""

from yoyo import step

__depends__ = {'__init__'}

steps = [
    step(
    """CREATE TABLE "boxes" (
	"id"	INTEGER PRIMARY KEY,
	"topic"	VARCHAR NOT NULL,
	"name"	VARCHAR,
	"hostname"	VARCHAR,
	"remark"	VARCHAR DEAFULT '',
	"first_seen"	timestamp DEFAULT CURRENT_TIMESTAMP,
	"online_since"	timestamp DEFAULT CURRENT_TIMESTAMP,
	"hw_type"	VARCHAR,
	"hw_version"	VARCHAR,
	"sw_type"	VARCHAR,
	"sw_version"	VARCHAR,
	"fallback_topic"	VARCHAR,
	"group_topic"	VARCHAR,
	"ip_address"	VARCHAR,
	"created_at"	timestamp DEFAULT CURRENT_TIMESTAMP,
	CONSTRAINT "topic_uc" UNIQUE("topic")
)"""
    ),(
    """DROP TABLE boxes"""
    ),
    step(
    """CREATE TABLE "events" (
	"id"	INTEGER PRIMARY KEY,
	"duration"	INTEGER DEFAULT 0,
	"water_amount"	INTEGER,
	"valve_id"	INTEGER NOT NULL REFERENCES "valves"("id") ON DELETE CASCADE,
	"created_at"	timestamp DEFAULT CURRENT_TIMESTAMP
)"""
    ),(
    """DROP TABLE events"""
    ),
    step(
    """CREATE TABLE "parameters" (
	"id"	INTEGER PRIMARY KEY,
	"key"	VARCHAR NOT NULL,
	"value"	VARCHAR,
	"created_at"	timestamp DEFAULT CURRENT_TIMESTAMP,
	CONSTRAINT "key_uc" UNIQUE("key")
)"""
    ),(
     """DROP TABLE parameters"""
    ),
    step(
    """CREATE TABLE "valves" (
	"id"	INTEGER PRIMARY KEY,
	"name"	VARCHAR,
	"remark"	VARCHAR DEFAULT '',
	"channel_nr"	INTEGER NOT NULL,
	"duration"	INTEGER DEFAULT 0,
	"interval"	INTEGER DEFAULT 4,
	"last_shedule"	timestamp DEFAULT '1970-01-01 00:00:00',
    "last_run"	timestamp DEFAULT '1970-01-01 00:00:00',
	"is_running"	INTEGER DEFAULT 0,
	"is_enabled"	INTEGER DEFAULT 0,
	"is_hidden"	INTEGER DEFAULT 1,
	"prio"	INTEGER DEFAULT 9,
	"box_id"	INTEGER NOT NULL REFERENCES "boxes"("id") ON DELETE CASCADE,
	"created_at"	timestamp DEFAULT CURRENT_TIMESTAMP,
	CONSTRAINT "channel_nr_box_id_uc" UNIQUE("channel_nr","box_id")
)"""
    ),(
    """DROP TABLE valves"""
    )
]



import aiohttp_jinja2
from aiohttp_session import get_session
from riego.db import get_db


class Dashboard():
    def __init__(self, app):
        self._db_conn = get_db().conn
        pass

    @aiohttp_jinja2.template('dashboard/index.html')
    async def index(self, request):
        session = await get_session(request)
        session['alert'] = {'class': 'alert-danger',
                            'heading': 'Danger!',
                            'text': 'lorem ipsum'}
        session.changed()

        c = self._db_conn.cursor()
        c.execute('SELECT * FROM valves')
        items = c.fetchall()
        self._db_conn.commit()

        return {'valves': items}



import logging
from logging.handlers import RotatingFileHandler
from pathlib import Path


def create_log(options):
    formatter = logging.Formatter("%(asctime)s;%(levelname)s; %(message)s ")

    stream_handler = logging.StreamHandler()
    # handler.setLevel(logging.DEBUG)
    stream_handler.setFormatter(formatter)

    logger = logging.getLogger("log")
    if options.verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    logger.addHandler(stream_handler)
    return logger


def create_event_log(options):
    Path(options.event_log).parent.mkdir(parents=True, exist_ok=True)
    formatter = logging.Formatter(
        "%(asctime)s %(message)s", "%Y-%m-%d %H:%M:%S")

    file_handler = RotatingFileHandler(options.event_log, mode='a',
                                       maxBytes=options.event_log_max_bytes,
                                       backupCount=options.event_log_backup_count,  # noqa: E501
                                       encoding=None, delay=0)

#    file_handler = logging.FileHandler(options.event_log)
    file_handler.setFormatter(formatter)

    logger = logging.getLogger("event_log")
    logger.setLevel(logging.DEBUG)
    logger.addHandler(file_handler)
    return logger



from sqlite3 import IntegrityError

from logging import getLogger
_log = getLogger(__name__)

_instance = None


def get_parameters():
    global _instance
    return _instance


def setup_parameters(app=None, options=None, db=None):
    global _instance
    if _instance is not None:
        del _instance
    _instance = Parameters(app=app, options=options, db=db)
    return _instance


class Parameters:
    def __init__(self, app=None, options=None, db=None):
        global _instance
        if _instance is None:
            _instance = self
        self._db_conn = db.conn

        self._start_time_1 = options.parameters_start_time_1
        self._max_duration = options.parameters_max_duartion

        self._load_all()

    @property
    def start_time_1(self):
        return self._start_time_1

    @start_time_1.setter
    def start_time_1(self, value):
        self._start_time_1 = value
        self._update_value_by_key(key="start_time_1", value=value)

    @property
    def max_duration(self):
        return self._max_duration

    @max_duration.setter
    def max_duration(self, value):
        self._max_duration = value
        self._update_value_by_key(key="max_duration", value=value)

    def _update_value_by_key(self, key=None, value=None):
        try:
            with self._db_conn:
                self._db_conn.execute(
                    'UPDATE parameters SET value = ? WHERE key = ?',
                    (value, key))
        except IntegrityError:
            pass
        try:
            with self._db_conn:
                self._db_conn.execute(
                    "INSERT INTO parameters (key, value) VALUES (?,?)",
                    (key, value))
        except IntegrityError:
            pass

    def _load_all(self):
        c = self._db_conn.cursor()
        c.execute("SELECT * from parameters")
        items = c.fetchall()
        self._db_conn.commit()
        for item in items:
            attr_name = "_" + item['key']
            if getattr(self, attr_name, None) is not None:
                setattr(self, attr_name, item['value'])
